#include <OneWire.h>                                                      // One-wire devices
#include <DallasTemperature.h>                                            // DS18B20
#include <SPI.h>
#include <Ethernet.h>

byte index;                                                               // Global loop index

// DS18B20
OneWire oneWire1(2);                                                      // Setup a oneWire instance to communicate with any OneWire device on pin 2
OneWire oneWire2(3);                                                      // Setup a oneWire instance to communicate with any OneWire device on pin 3
DallasTemperature temperatureSensors1(&oneWire1);                         // Pass oneWire reference to DallasTemperature library
DallasTemperature temperatureSensors2(&oneWire2);                         // Pass oneWire reference to DallasTemperature library
DeviceAddress sensorAddress[30];                                          // variable to hold sensor addresses
float sensorNewTempC[30];                                                 // variable to hold sensor temperature
float sensorOldTempC[30];                                                 // variable to hold sensor temperature
byte sensorTimeCounter[30];                                               // variable to hold seconds without temp being sent by UDP
byte sensorCount;                                                         // variable to hold number of sensors
byte sensorindex;                                                         // variable to hold sensor pointer
byte changes;                                                             // How many sensors changed this loop
long starttime;                                                           // variable to measure time used for a part of code
long interval1;
long interval2;
long interval3;
long interval4;
long interval5;
long interval6;


// Ethernet
byte mac[] = {0x00, 0xAA, 0xBB, 0xCC, 0xDE, 0x02};                        // MAC address of the ethernet shield
byte macBuffer[6];                                                        // create a buffer to hold the MAC address
IPAddress ip_receiver(255,255,255,255);                                   // 192.168.1.104-laptop FOR TESTING
//IPAddress gateway(192,168,1,1);                                          // Not required?
//IPAddress subnet(255,255,0,0);                                           // Not required?
unsigned int receiverport = 8888;                                         // Port to send packets to
int packetSize;                                                           // Size of received packet
char packetBuffer[1];                                                     // Buffer for incoming packets
EthernetUDP UDP;                                                          // An EthernetUDP instance to let us send and receive packets over UDP


void setup(void)
{
// Serial communication (USB)
  Serial.begin(115200);
  while (!Serial) {;}                                                     // wait for serial port to connect. Needed for native USB port only
  Serial.println(F("Serial port active"));
  
// DS18B20
  temperatureSensors1.begin();                                            // Start up the library
  temperatureSensors2.begin();                                            // Start up the library
  temperatureSensors1.requestTemperatures();                              // Send command to all the sensors for temperature conversion
  temperatureSensors2.requestTemperatures();                              // Send command to all the sensors for temperature conversion
  getSensorsOnBus();                                                      // Find all sensors on the bus
  for (index = 0; index < 8; index++) {sensorOldTempC[index]=-55;}        // Fill up array with -55 to force sendTempUDP on first pass

// Ethernet
  Serial.println(F("Ethernet.begin(mac)"));
  Ethernet.begin(mac);
  Serial.println(F("Udp.begin(receiverport)"));
  UDP.begin(receiverport);

// Ethernet connection:
  Serial.println(F("Check Ethernet connection:"));
  if (Ethernet.begin(mac) == 0)
  {
    Serial.println(F("Failed to configure Ethernet using DHCP"));
    if (Ethernet.hardwareStatus() == EthernetNoHardware)
    {
      Serial.println(F("Ethernet shield was not found. :("));
    }
    else if (Ethernet.linkStatus() == LinkOFF)
    {
      Serial.println(F("Ethernet cable is not connected."));
    }
    while (true) {delay(1);}                                              // no point in carrying on, so do nothing forevermore:
  }
  Serial.print(F("My IP address: "));                                     // print your local IP address:
  Serial.println(Ethernet.localIP());
  Ethernet.MACAddress(macBuffer);                                         // Put MAC address in macBuffer
  Serial.print(F("The MAC address is: "));
  for (index = 0; index < 6; index++)                                     // Read out the MAC
  {
    if (macBuffer[index]<16) {Serial.print('0');}                         // Formatting
    Serial.print(macBuffer[index], HEX);
    if (index < 5) {Serial.print('-');}                                   // Formatting
  }
  Serial.println();
  Serial.println(F("Program starting now."));
  Serial.println();
}


//===============================================================================================================================================
void loop(void)
{
  changes=0;
starttime=millis();
  for (sensorindex = 0;  sensorindex < sensorCount;  sensorindex++)
  {
    if (sensorTimeCounter[sensorindex]==30)                               // 30 loops passed without sending this sensor's temperature
    {
      sendTempUDP(sensorAddress[sensorindex]);                            // Send temperature at least once evey 30 conversions
    }
    else if (sensorNewTempC[sensorindex]!=sensorOldTempC[sensorindex])
    {
      changes++;
      sendTempUDP(sensorAddress[sensorindex]);                            // Send temperature when changed
    }
    else
    {
      sensorTimeCounter[sensorindex]++;                                   // Increase the sensorTimeCounter for this sensor
    }
  }
interval1=millis()-starttime;
starttime=millis();
  sendTempSerial();                                                       // Print sensor data for testingstarttime=millis();
interval2=millis()-starttime;
starttime=millis();
  while (oneWire1.read() == 0) {  }                                       // Wait for conversion end
interval3=millis()-starttime;
starttime=millis();
  for (sensorindex = 0;  sensorindex < sensorCount;  sensorindex++)       // Read out all temperatures (6 sensors = 378ms)
    {
      sensorNewTempC[sensorindex] = temperatureSensors1.getTempCByIndex(sensorindex);
    }
interval4=millis()-starttime;
starttime=millis();
  temperatureSensors1.requestTemperatures();                               // Send command to all the sensors for temperature conversion, 6 sensors 524ms
interval5=millis()-starttime;
starttime=millis();
Serial.print(F("--"));
Serial.print(changes);
Serial.print(F("--  "));
if (interval1<10) Serial.print(" ");
if (interval1<100) Serial.print(" ");
Serial.print(interval1);
Serial.print("  ");
if (interval2<10) Serial.print(" ");
if (interval2<100) Serial.print(" ");
Serial.print(interval2);
Serial.print("  ");
if (interval3<10) Serial.print(" ");
if (interval3<100) Serial.print(" ");
Serial.print(interval3);
Serial.print("  ");
if (interval4<10) Serial.print(" ");
if (interval4<100) Serial.print(" ");
Serial.print(interval4);
Serial.print("  ");
if (interval5<10) Serial.print(" ");
if (interval5<100) Serial.print(" ");
Serial.println(interval5);
  packetSize = UDP.parsePacket();
  if(packetSize)
  {
    Serial.print(F("Received packet of size "));
    Serial.println(packetSize);
    UDP.read(packetBuffer,1);
    if(packetBuffer[0]=='R')
    {
    Serial.println(packetBuffer);
    temperatureSensors1.begin();                                            // Start up the library again
    getSensorsOnBus();
    }
    Serial.println();
  }
}
//===============================================================================================================================================



void getSensorsOnBus()                                                    // Find sensors on the bus
{
  Serial.print(F("Locating devices..."));
  sensorCount=0;
  do                                                                      // Loop here when no sensors are found
  {
    sensorCount = temperatureSensors1.getDeviceCount();
    delay(100);                                                           // Wait 100ms
  } while (sensorCount==0);
  Serial.print(F("Found "));
  Serial.print(sensorCount, DEC);
  Serial.println(F(" device(s)."));
  for (index = 0;  index < sensorCount;  index++)                         // Fill up the sensorAddress array with addresses found
  {
    temperatureSensors1.getAddress(sensorAddress[index], index);           // Why twice index??
    temperatureSensors1.setResolution(sensorAddress[index], 12);           // Default is 12 bits (0,0625)
    sensorTimeCounter[index]=0;                                            // Reset sensorTimeCounter
    printAddress(sensorAddress[index]);
    Serial.print(F("   "));
    sensorNewTempC[index]=temperatureSensors1.getTempCByIndex(index);
    sensorOldTempC[index]=sensorNewTempC[index];
    Serial.println(sensorNewTempC[index]);
  }
  Serial.println();
}


void printAddress(DeviceAddress sensorAddress)                            // Function to print all sensor addresses
{
  for (byte i = 0; i < 8; i++)
  {
    if (sensorAddress[i] < 16) Serial.print("0");
    Serial.print(sensorAddress[i], HEX);
  }
  Serial.print("  ");
}


void sendTempUDP(DeviceAddress sensorAddress)                             // Routine to print all sensor addresses
{
  sensorTimeCounter[sensorindex]=0;                                       // Reset sensorTimeCounter
  UDP.beginPacket(ip_receiver, receiverport);
  for (byte i = 0; i < 8; i++)
  {
    if (sensorAddress[i] < 16) {UDP.print('0');}                          // Leading 0
    UDP.print(sensorAddress[i], HEX);
  }
  UDP.print('=');
  UDP.print(sensorNewTempC[sensorindex], 1);
  UDP.print("  ");
  UDP.endPacket();
  sensorOldTempC[sensorindex]=sensorNewTempC[sensorindex];
}


void sendTempSerial()
{
  sensorTimeCounter[sensorindex]=0;                                       // Reset sensorTimeCounter
  for (sensorindex = 0;  sensorindex<sensorCount; sensorindex++)
  {
    Serial.print(sensorindex);
    Serial.print(F(" "));
    Serial.print(sensorNewTempC[sensorindex],2);
    Serial.print(F(" ("));
    if (sensorTimeCounter[sensorindex]<10) {Serial.print(F(" "));}           // For allignment
    Serial.print(sensorTimeCounter[sensorindex]);
    Serial.print(F(")   "));
  }
//  Serial.println();
}
